@@ -1,6 +1,8 @@
 #include <pybind11/pybind11.h>
 
 #include "Models/ModelTypes.hpp"
+#include "Models/DoubleModel.hpp"
+#include "Models/SpdModel.hpp"
 #include "Models/Policies/PriorPolicy.hpp"
 #include "Models/PosteriorSamplers/PosteriorSampler.hpp"
 #include "cpputil/Ptr.hpp"
@@ -16,6 +18,24 @@ namespace BayesBoom {
     py::class_<Model, Ptr<Model>>(boom, "Model")
         ;
 
+    py::class_<DoubleModel, Model, Ptr<DoubleModel>>(boom, "DoubleModel", py::multiple_inheritance())
+        .def("logp", [](const DoubleModel &m, double x) {
+          return m.logp(x);
+        },
+          py::arg("x"),
+          "The log density evaluated at 'x'.")
+        ;
+
+    py::class_<SpdModel, Model, Ptr<SpdModel>>(
+        boom, "SpdModel", py::multiple_inheritance())
+        .def("logp",
+             [](const SpdModel &model, const SpdMatrix &x) {
+               return model.logp(x);
+             },
+             py::arg("x"),
+             "The log density evaluated at 'x'.")
+        ;
+
     py::class_<PosteriorSampler, Ptr<PosteriorSampler>>(
         boom, "PosteriorSampler")
         .def("draw", &PosteriorSampler::draw)
@@ -36,6 +56,8 @@ namespace BayesBoom {
              "performed by any posterior samplers that have been assigned \n"
              "to this model by  'set_method'.\n")
         ;
+
+
   }  // Module
 
 }  // namespace BOOM
